import sqlite3
from pathlib import Path
import threading
import json
lock = threading.Lock()



class BaseDB:
    def __init__(self,db_path: str,db_name: str,create_table_sql: str):

        self.db_path = Path(db_path)
        self.table = db_name
        self.create_table_sql = create_table_sql
        self.load()

        table_columns = self.db.execute(f"PRAGMA table_info({self.table})")
        self.table_columns = [col['name'] for col in table_columns]

    def load(self):
        self.db= SQLiteDB(self.db_path,self.table,self.create_table_sql)

    def save(self):
        self.db.save()

    def clear(self):
        self.db.clear()

    def buffer_clear(self):
        self.buffer=[]

    def add(self,data_list: list):
        if data_list == []:
            return

        keys = list(data_list[0].keys())
        keys = [k for k in keys if k in self.table_columns]
        columns = ','.join(keys)
        placeholders = ','.join('?' * len(keys))
        sql = f'INSERT OR IGNORE INTO {self.table} ({columns}) VALUES ({placeholders})'
        values_list = [tuple(data[k] for k in keys) for data in data_list]
        self.db.execute_batch(sql, values_list)

    def update(self,data_list: list):
        if data_list == []:
            return
        keys = list(data_list[0].keys())
        keys = [k for k in keys if k in self.table_columns]
        set_clause = ','.join([f"{k} = ?" for k in keys])
        sql = f"UPDATE {self.table} SET {set_clause} WHERE id = ?"
        values_list = [tuple(data[k] for k in keys) + (data['id'],) for data in data_list]
        self.db.execute_batch(sql, values_list)
    
    def update_key(self,key: str,value):
        sql = f"UPDATE {self.table} SET {key} = ?"
        self.db.execute(sql, (value,))

    def search(self,key: str,value):
        sql = f"SELECT * FROM {self.table} WHERE {key} = ?"
        return self.db.execute(sql, (value,))

    def search_all(self):
        sql = f"SELECT * FROM {self.table}"
        return self.db.execute(sql)
    
class SQLiteDB:
    def __init__(self,db_path,table,create_table_sql):
        self.db_path = Path(db_path)
        self.table = table
        self.create_table_sql = create_table_sql
        self.load()
    
    def load(self):
        if not self.db_path.parent.exists():
            self.db_path.parent.mkdir(parents=True)
        if not self.db_path.exists():
            self.db_path.touch()

        with lock:
            self.conn = sqlite3.connect(self.db_path,check_same_thread=False)
            self.conn.row_factory = sqlite3.Row
            self.cursor = self.conn.cursor()
            self.cursor.execute(self.create_table_sql)
            self.conn.commit()

    def save(self):
        ...
    
    def clear(self):
        with lock:
            self.cursor.execute(f"DROP TABLE IF EXISTS {self.table}")
            self.conn.commit()
        self.load()

    def execute(self,SQL,values=()):
        with lock:
            self.cursor.execute(SQL,values)
            self.conn.commit()
            res=self.cursor.fetchall()
        res=[dict(row) for row in res]
        return res

    def execute_batch(self,SQL,values_list):
        with lock:
            self.cursor.executemany(SQL,values_list)
            self.conn.commit()


import yaml

class BaseDataclass:
    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)
    @classmethod
    def from_yaml(cls, path: str):
        with open(path, "r") as f:
            return cls.from_dict(yaml.safe_load(f))



import hashlib
import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

def hash_str(s: str):
    return hashlib.md5(s.encode()).hexdigest()

def executor(task_func,task_list,num_threads=1):
    if num_threads<=1:
        for task in tqdm(task_list):
            task_func(task)
    else:
        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = [executor.submit(task_func,task) for task in task_list]
            pbar = tqdm(as_completed(futures), total=len(futures))
            for future in pbar:
                future.result()


from functools import wraps
from ..database.sqlite_db import BaseDB
from pathlib import Path
import hashlib
import json
import inspect

class CacheDB(BaseDB):
    def __init__(self,cache_path: str,cache_name: str):
        cache_path=Path(cache_path)
        db_path= cache_path / f'{cache_name}.db'
        table='cache'
        create_table_sql = \
            f"""
            CREATE TABLE IF NOT EXISTS {table} (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                hash TEXT UNIQUE,
                input TEXT,
                output TEXT
            );
            """
        super().__init__(db_path,table,create_table_sql)

    def search_cache(self,input_hash):
        res_json = self.db.execute(f'SELECT output FROM {self.table} WHERE hash = ?',(input_hash,))
        if len(res_json) == 0:
            return None
        return res_json[0]['output']
    
    def update_cache(self,input_hash,input_json,output_json):
        self.db.execute(f'INSERT OR IGNORE INTO {self.table} (hash,input,output) VALUES (?,?,?)',(input_hash,input_json,output_json))


class Cache:
    def __init__(self,cache_dir='cache',cache_name='cache'):
        self.cache_dir = cache_dir
        self.cache_name = cache_name
        self.cache_db = CacheDB(self.cache_dir,self.cache_name)
    
    def input_encode(self,inp):
        inp_json = json.dumps(inp,ensure_ascii=False)
        inp_hash = hashlib.md5(inp_json.encode('utf-8')).hexdigest()
        inp_encoded = {
            'hash': inp_hash,
            'json': inp_json,
        }
        return inp_encoded

    def output_encode(self,out):
        out_encoded = json.dumps(out,ensure_ascii=False)
        return out_encoded
    
    def output_decode(self,out_encoded):
        out = json.loads(out_encoded)
        return out

    def find_cache(self,inp_encoded):
        inp_hash = inp_encoded['hash']
        out_encoded = self.cache_db.search_cache(inp_hash)

        if out_encoded is None:
            out = None
        else:
            out = self.output_decode(out_encoded)
        return out
    
    def save_cache(self,inp_encoded,out):
        inp_hash = inp_encoded['hash']
        inp_json = inp_encoded['json']
        out_encoded = self.output_encode(out)
        self.cache_db.update_cache(inp_hash,inp_json,out_encoded)   

    def __call__(self,func):
        @wraps(func)
        def wrapper(*args,**kwargs):
            sig = inspect.signature(func)
            bound_args = sig.bind(*args,**kwargs)
            bound_args.apply_defaults()

            inp=bound_args.arguments.copy()
            inp.pop('self',None)
            inp.pop('cls',None)
            
            inp_encoded = self.input_encode(inp)
            out = self.find_cache(inp_encoded)

            if out is None:
                out = func(*args,**kwargs)
                self.save_cache(inp_encoded,out)
            return out
    
        return wrapper


from openai import OpenAI
from ..cache import Cache

class LLM:
    def __init__(self,
                api_key,
                base_url,
                default_model_args = None,
                use_cache = True,
                cache_dir='cache/OpenAIAPI/',
                cache_name='llm_cache'
                ):
        
        self.api_key = api_key
        self.base_url = base_url

        self.client = OpenAI(
            api_key = self.api_key, 
            base_url = self.base_url,
        )
        self.default_model_args = default_model_args if default_model_args is not None else {}

        self.use_cache = use_cache
        if self.use_cache:
            self.cache = Cache(cache_dir,cache_name)

    def generate(self,messages , model_args = None,**kwargs):
        if model_args is None:
            model_args = self.default_model_args

        if self.use_cache:
            input={'messages':messages,'model_args':model_args,'kwargs':kwargs}
            inp_encoded = self.cache.input_encode(input)
            response = self.cache.find_cache(inp_encoded)
            if response is not None:
                return response

        response={}        
        try:
            completion = self.client.chat.completions.create(
                **model_args,
                messages=messages,
            )
            response['status']=1
            response['answer']=completion.choices[0].message.content
        except Exception as e:
            response['status']=0
            response['answer']= str(e)

        try:
            response['usage_prompt_tokens']=completion.usage.prompt_tokens
            response['usage_completion_tokens']=completion.usage.completion_tokens
            response['usage_total_tokens']=completion.usage.total_tokens
        except:
            response['usage_prompt_tokens']=None
            response['usage_completion_tokens']=None
            response['usage_total_tokens']=None

        try:
            response['usage_cached_tokens']=completion.usage.prompt_tokens_details.cached_tokens
        except:
            response['usage_cached_tokens']=None

        if self.use_cache:
            self.cache.save_cache(inp_encoded,response)

        return response

class Embedding:
    def __init__(self,
                api_key,
                base_url,
                default_model_args = None,
                use_cache = True,
                cache_dir='cache/OpenAIAPI/',
                cache_name='emb_cache'
                ):
        
        self.api_key = api_key
        self.base_url = base_url

        self.client = OpenAI(
            api_key = self.api_key, 
            base_url = self.base_url,
        )
        self.default_model_args = default_model_args if default_model_args is not None else {}

        self.use_cache = use_cache
        if self.use_cache:
            self.cache = Cache(cache_dir,cache_name)
    
    def generate(self,text,model_args = None,**kwargs):
        if model_args is None:
            model_args = self.default_model_args

        if self.use_cache:
            input={'text':text,'model_args':model_args,'kwargs':kwargs}
            inp_encoded = self.cache.input_encode(input)
            response = self.cache.find_cache(inp_encoded)
            if response is not None:
                return response

        response={}
        try:
            embedding = self.client.embeddings.create(
                **model_args,
                input=text,
                )
            response['status']=1
            response['embedding']=embedding.data[0].embedding
        except Exception as e:
            response['status']=0
            response['embedding']=str(e)
        
        if self.use_cache:
            self.cache.save_cache(inp_encoded,response)
        
        return response


from functools import wraps
import concurrent.futures
import time

class Retry:
    def __init__(self,max_attempt: int=5,wait: float=0.1,timeout: float=10000):
        self.max_attempt = max_attempt
        self.wait = wait
        self.timeout = timeout

    def __call__(self,func: callable):
        @wraps(func)
        def wrapper(*args,**kwargs):
            
            for attempt in range(1,self.max_attempt+1):
                kwargs_with_attempt = {**kwargs, 'attempt': attempt}
                try:
                    with concurrent.futures.ThreadPoolExecutor() as executor:
                        future = executor.submit(func, *args, **kwargs_with_attempt)
                        result = future.result(timeout=self.timeout)
                    return result
                except Exception as e:
                    time.sleep(self.wait)

            return None
        
        return wrapper


import faiss
import sqlite3
from pathlib import Path
import threading
import numpy as np
import json
lock = threading.Lock()

class BaseVDB:
    def __init__(self,vdb_path: str,vdb_name: str,vdb_dim: int):
        self.vdb_path = Path(vdb_path)
        self.vdb_name = vdb_name
        self.vdb_dim = vdb_dim
        self.load()
        
        self.buffer=[]

    def load(self):
        self.vdb_file_path = self.vdb_path / f'{self.vdb_name}.vdb'
        self.vdb = FassiVDB(self.vdb_file_path,self.vdb_dim)

    def save(self):
        self.vdb.save()

    def clear(self):
        self.vdb.clear()
        self.load()

    def buffer_clear(self):
        self.buffer=[]

    def vdb_add(self,data_list: list):
        if data_list == []:
            return
        ids = [data['id'] for data in data_list]
        vectors = np.array([json.loads(data['embedding']) for data in data_list])
        self.vdb.add(ids,vectors)
    
    def vdb_search(self,vector: list,topk: int=10):        
        vector = np.array(vector).reshape(1, -1)
        distances,ids = self.vdb.search(vector, topk)
        if ids is None:
            return []
        pairs = [(float(d), int(i)) for d, i in zip(distances[0], ids[0]) if i != -1]

        res=[{'distance':p[0],'id':p[1]} for p in pairs]
        return res

class FassiVDB:
    def __init__(self,vdb_path,vdb_dim):
        self.vdb_path = Path(vdb_path)
        self.vdb_dim = vdb_dim
        self.load()

    def load(self):
        if self.vdb_path.exists():
            self.vdb = faiss.read_index(str(self.vdb_path))
        elif self.vdb_dim is not None:
            self.vdb = faiss.IndexIDMap(faiss.IndexFlatL2(self.vdb_dim))
            self.save()
        else:
            self.vdb = None
        
    def save(self):
        if self.vdb is None:
            return
        with lock:
            if not self.vdb_path.parent.exists():
                self.vdb_path.parent.mkdir(parents=True)
            faiss.write_index(self.vdb, str(self.vdb_path))

    def clear(self):
        if self.vdb is None:
            return
        file_path = Path(self.vdb_path)
        if file_path.exists():
            file_path.unlink()
        self.load()

    def add(self,ids,items):
        if self.vdb is None:
            return
        with lock:
            self.vdb.add_with_ids(items,np.array(ids,dtype=np.int64))
    
    def search(self,item,topk):
        if self.vdb is None:
            return None,None
        with lock:
            distances, ids = self.vdb.search(item, topk)

        return distances,ids
